# copyright (C) 1997-1999 Jean-Luc Fontaine (mailto:jfontain@free.fr)
# this program is free software: please read the COPYRIGHT file enclosed in this package or use the Help Copyright menu

set rcsId {$Id: blt2d.tcl,v 1.34.1.6 2000/05/28 14:24:15 jfontain Exp $}

class blt2DViewer {

    proc blt2DViewer {this path} viewer {} {
        $path configure -cursor {} -plotpadx 2 -plotpady 2   ;# use minimum padding for extreme values, flat zero line to be visible
        $path yaxis configure -tickshadow {} -title {} -tickfont $font::(smallNormal)
        $path legend configure -borderwidth 1 -font $font::(mediumNormal) -activebackground white

        viewer::setupDropSite $this $path                                                            ;# allow dropping of data cells

        set blt2DViewer::($this,elements) {}                                             ;# initialize list of data line identifiers
        set blt2DViewer::($this,colorIndex) 0
        set blt2DViewer::($this,path) $path
    }

    proc ~blt2DViewer {this} {
        if {[info exists blt2DViewer::($this,drag)]} {
            delete $blt2DViewer::($this,drag)
        }
        eval delete $blt2DViewer::($this,elements)                                                       ;# delete existing elements
        if {[info exists blt2DViewer::($this,selector)]} {
            delete $blt2DViewer::($this,selector)
        }
    }

    proc supportedTypes {this} {
        return {integer real}
    }

    proc dragData {this format} {
        set selectedElements [selector::selected $blt2DViewer::($this,selector)]
        switch $format {
            OBJECTS {
                if {[llength $selectedElements]>0} {
                    return $selectedElements                                            ;# return selected elements if there are any
                } elseif {[llength $blt2DViewer::($this,elements)]==0} {
                    return $this                                                   ;# return graph itself if it contains no elements
                } else {
                    return {}                                                                            ;# return nothing otherwise
                }
            }
            DATACELLS {
                return [cellsFromElements $this $selectedElements]
            }
        }
    }

    proc validateDrag {this x y} {
        if {[llength $blt2DViewer::($this,elements)]==0} {
            return 1                                                                                   ;# allow drag of empty viewer
        }
        # allow dragging if only from a selected cell
        return [expr {\
            [lsearch -exact [selector::selected $blt2DViewer::($this,selector)] [$blt2DViewer::($this,path) legend get @$x,$y]]>=0\
        }]
    }

    proc monitorCell {this array row column} {
        viewer::registerTrace $this $array
        set cell ${array}($row,$column)
        if {[lsearch -exact [cellsFromElements $this $blt2DViewer::($this,elements)] $cell]>=0} return     ;# already charted, abort
        set label [viewer::label $array $row $column]
        set element [newElement $this $blt2DViewer::($this,path)\
            -label $label -color [lindex $global::viewerColors $blt2DViewer::($this,colorIndex)]\
        ]
        # keep track of element existence
        switched::configure $element -deletecommand "blt2DViewer::deletedElement $this $array $element"
        # circle through colors
        set blt2DViewer::($this,colorIndex) [expr {($blt2DViewer::($this,colorIndex)+1)%[llength $global::viewerColors]}]
        lappend blt2DViewer::($this,elements) $element                                                     ;# register new data line
        set blt2DViewer::($this,cell,$element) $cell
        if {[info exists blt2DViewer::($this,selector)]} {                          ;# selector may not exist if dragging disallowed
            selector::add $blt2DViewer::($this,selector) $element
        }
        if {[string first ? $label]>=0} {                                                          ;# label cannot be determined yet
            set blt2DViewer::($this,relabel,$element) {}
        }
    }

    proc cells {this} {
        return [cellsFromElements $this $blt2DViewer::($this,elements)]
    }

    proc deletedElement {this array element} {
        viewer::unregisterTrace $this $array                                          ;# trace may no longer be needed on this array
        ldelete blt2DViewer::($this,elements) $element
        if {[info exists blt2DViewer::($this,selector)]} {                          ;# selector may not exist if dragging disallowed
            selector::remove $blt2DViewer::($this,selector) $element
        }
        unset blt2DViewer::($this,cell,$element)
    }

    proc update {this array args} {                                                               ;# update display using cells data
        updateTimeDisplay $this [set seconds [clock seconds]]
        foreach element $blt2DViewer::($this,elements) {
            set cell $blt2DViewer::($this,cell,$element)
            if {[string first $array $cell]<0} continue                                  ;# check that cell belongs to updated array
            if {[info exists $cell]} {
                if {[info exists blt2DViewer::($this,relabel,$element)]} {                 ;# if label is not yet defined, update it
                    viewer::parse $cell array row column type
                    set label [viewer::label $array $row $column]
                    switched::configure $element -label $label
                    if {[string first ? $label]<0} {                                                 ;# label now completely defined
                        unset blt2DViewer::($this,relabel,$element)
                    }
                }
                updateElement $this $element $seconds [set $cell]                      ;# may be ? if cell value is meant to be void
            } else {
                updateElement $this $element $seconds ?                                                     ;# data no longer exists
            }
        }
    }

    virtual proc newElement {this path args}                                       ;# let derived class create an element of its own

    virtual proc updateElement {this element seconds value}      ;# let derived class (such as graph, bar chart, ...) update element

    virtual proc updateTimeDisplay {this seconds} {}        ;# eventually let derived class (such as graph) update axis, for example

    proc cellsFromElements {this elements} {
        set cells {}
        foreach element $elements {
            lappend cells $blt2DViewer::($this,cell,$element)
        }
        return $cells
    }

    proc setElementsState {this elements select} {
        if {$select} {
            set action activate
        } else {
            set action deactivate
        }
        set path $blt2DViewer::($this,path)
        foreach element $elements {
            $path legend $action $element
        }
    }

    proc setSelection {this x y} {
        if {[string length [set element [$blt2DViewer::($this,path) legend get @$x,$y]]]>0} {                         ;# in a legend
            selector::select $blt2DViewer::($this,selector) $element
        }
    }

    proc toggleSelection {this x y} {
        if {[string length [set element [$blt2DViewer::($this,path) legend get @$x,$y]]]>0} {                         ;# in a legend
            selector::toggle $blt2DViewer::($this,selector) $element
        }
    }

    proc extendSelection {this x y} {
        if {[string length [set element [$blt2DViewer::($this,path) legend get @$x,$y]]]>0} {                         ;# in a legend
            selector::extend $blt2DViewer::($this,selector) $element
        }
    }

    proc allowDrag {this} {
        set path $blt2DViewer::($this,path)

        set blt2DViewer::($this,drag) [new dragSite -path $path -validcommand "blt2DViewer::validateDrag $this"]
        dragSite::provide $blt2DViewer::($this,drag) OBJECTS "blt2DViewer::dragData $this"                ;# provide list of objects
        dragSite::provide $blt2DViewer::($this,drag) DATACELLS "blt2DViewer::dragData $this"

        set blt2DViewer::($this,selector) [new objectSelector -selectcommand "blt2DViewer::setElementsState $this"]
        bind $path <ButtonRelease-1> "blt2DViewer::setSelection $this %x %y"
        bind $path <Control-ButtonRelease-1> "blt2DViewer::toggleSelection $this %x %y"
        bind $path <Shift-ButtonRelease-1> "blt2DViewer::extendSelection $this %x %y"
    }

}
