# copyright (C) 1997-2000 Jean-Luc Fontaine (mailto:jfontain@free.fr)
# this program is free software: please read the COPYRIGHT file enclosed in this package or use the Help Copyright menu

set rcsId {$Id: module.tcl,v 1.2.1.5 2000/06/13 19:38:27 jfontain Exp $}


class module {

    proc module {this name index} {       ;# index may be forced (when coming from a save file), left empty for automatic generation
        if {[string length $index]==0} {
            set index [newIndex $name]
        } else {                                                                                    ;# index was passed as parameter
            addIndex $name $index
        }
        set module::($this,name) $name
        set module::($this,index) $index
    }

    proc ~module {this} {
        if {[info exists module::($this,interpreter)]} {
            if {[procedureExists $this terminate]} {
                $module::($this,interpreter) eval $module::($this,name)::terminate ;# invoke module terminate procedure if it exists
            }
            interp delete $module::($this,interpreter)
        }
        if {[info exists module::($this,namespace)]} {
            namespace delete ::$module::($this,namespace)
        }
        deleteIndex $this
    }

    proc newIndex {name} {              ;# find a new index for the module, by eventually finding a hole in that module indices list
        variable indices

        if {![info exists indices($name)]} {
            set indices($name) {}
        }
        set new 0
        foreach index $indices($name) {
            if {$index!=$new} break                                                                                  ;# found a hole
            incr new
        }
        set indices($name) [lsort -integer [concat $indices($name) $new]]
        return $new
    }

    proc addIndex {name index} {
        variable indices

        if {![info exists indices($name)]} {
            set indices($name) $index
            return
        }
        if {[lsearch -exact $indices($name) $index]>=0} {
            error "trying to add an existing index: $index"
        }
        set indices($name) [lsort -integer [concat $indices($name) $index]]
    }

    proc deleteIndex {this} {
        variable indices

        set name $module::($this,name)
        ldelete indices($name) $module::($this,index)
        if {[llength $indices($name)]==0} {
            unset indices($name)
        }
    }

    proc load {this} {              ;# load a module in its own interpreter, in order to allow multiple instances of the same module
        set name $module::($this,name)
        set index $module::($this,index)
        if {$index==0} {                                                                               ;# first instance of a module
            set namespace $name                                                                          ;# use original module name
        } else {
            set namespace ${name}<$index>                                          ;# this is another instance of an existing module
        }
        set interpreter [interp create]                                                ;# use a separate interpreter for each module
        set module::($this,interpreter) $interpreter
        $interpreter eval {                    ;# since Tk is not loaded in module interpreter, provide a background error procedure
            proc bgerror {error} {
                puts stderr $::errorInfo
                exit 1
            }
        }
        $interpreter eval "
            set auto_path [list $::auto_path]                          ;# copy a few required global variables in module interpreter
            set ::packageDirectory($name) $::packageDirectory($name)
            package require $name                                                                           ;# module is loaded here
        "
        # we never get here if there is an error when the module is loaded
        set module::($this,namespace) $namespace
        # the new namespace, "interface" to the protected module namespace in its interpreter, is child of the global namespace
        namespace eval ::$namespace [subst -nocommands {proc update {} {$interpreter eval ::${name}::update}}]
        set module::($this,initialize) [procedureExists $this initialize]
        if {$module::($this,initialize)} {                                                            ;# initialize procedure exists
            namespace eval ::$namespace [subst -nocommands {        ;# create an interface initialize procedure within new namespace
                proc initialize {arguments} {                     ;# arguments are a list of option / value (eventually empty) pairs
                    $interpreter eval "
                        array set _options [list \$arguments]
                        ::${name}::initialize _options
                        unset _options
                    "
                }
            }]
        }
        set module::($this,version) [$interpreter eval "package provide $name"]
        synchronize $this                                                ;# initialize namespace data from module in its interpreter
        validateColumnTitles $this
        # keep on eye on special module data array member "update"
        $interpreter alias _updated ::module::updated $this
        $interpreter eval "trace variable ::${name}::data(updates) w _updated"
        # setup interface to messenger:
        $interpreter alias pushMessage ::lifoLabel::push $::global::messenger
        $interpreter alias popMessage ::lifoLabel::pop $::global::messenger
        $interpreter alias flashMessage ::lifoLabel::flash $::global::messenger
    }

    proc updated {this args} {                                 ;# module data was just updated. ignore already known trace arguments
        set namespace $module::($this,namespace)
        lifoLabel::push $global::messenger "$namespace data update..."
        update idletasks
        foreach name [array names ::${namespace}::data {[0-9]*,[0-9]*}] {        ;# first clear dynamic data (so rows can disappear)
            unset ::${namespace}::data($name)
        }
        synchronize $this {[0-9]*,[0-9]*}                                            ;# just copy all dynamic data from module array
        # copy updates counter
        set ::${namespace}::data(updates) [$module::($this,interpreter) eval "set ::$module::($this,name)::data(updates)"]
        lifoLabel::pop $global::messenger
    }

    proc clear {this} {                                                         ;# act as if module code was updated with empty data
        set namespace $module::($this,namespace)
        foreach name [array names ::${namespace}::data {[0-9]*,[0-9]*}] {         ;# clear dynamic data (so rows can disappear) then
            unset ::${namespace}::data($name)
        }
        # copy updates counter
        set ::${namespace}::data(updates) [$module::($this,interpreter) eval "set ::$module::($this,name)::data(updates)"]
    }

    proc synchronize {this {pattern *}} {                       ;# copy data from module in its interpreter to module namespace here
        array set ::$module::($this,namespace)::data [$module::($this,interpreter) eval\
            "array get ::$module::($this,name)::data {$pattern}"]
    }

    proc validateColumnTitles {this} {
        foreach {name label} [array get ::$module::($this,namespace)::data *,label] {
            if {[string first ? $label]>=0} {
                scan $name %u column
                puts stderr "in $module::($this,namespace) module, column $column label contains a ? character: \"$label\""
                exit 1
            }
        }
    }

    proc procedureExists {this name} {                     ;# see if procedure exists in module namespace within its own interpreter
        return [$module::($this,interpreter) eval\
            [subst -nocommands {expr {[string length [namespace eval ::$module::($this,name) {info proc $name}]]>0}}]\
        ]
    }

}
