# copyright (C) 1997-2000 Jean-Luc Fontaine (mailto:jfontain@free.fr)
# this program is free software: please read the COPYRIGHT file enclosed in this package or use the Help Copyright menu

set rcsId {$Id: modules.tcl,v 1.19.1.18 2000/06/17 20:12:44 jfontain Exp $}


class modules {

    class instance {

        proc instance {this module index} {
            set modules::instance::($this,module) $module
            set modules::instance::($this,loaded) [new module $module $index]
        }

        proc ~instance {this} {
            delete $modules::instance::($this,loaded)
        }

        proc load {this} {                               ;# load module since its switches need to be known for command line parsing
            set loaded $modules::instance::($this,loaded)
            module::load $loaded
            set namespace $module::($loaded,namespace)
            set modules::instance::($this,namespace) $namespace
            catch {set modules::instance::($this,switches) [set ::${namespace}::data(switches)]}       ;# module may take no options
            set modules::instance::($this,initialize) $module::($loaded,initialize)
            set modules::instance::($this,version) $module::($loaded,version)
            initialize $this
        }

        proc initialize {this} {
            set namespace $modules::instance::($this,namespace)
            if {![catch {set ${namespace}::data(identifier)} identifier]} {            ;# store identifier if it exists and is valid
                if {![modules::validName $identifier]} {
                    foreach {name index} [modules::decoded $namespace] {}
                    puts stderr "\"$name\" module identifier: \"$identifier\" contains invalid characters"
                    exit 1
                }
                set modules::instance::($this,identifier) $identifier
            }
            catch {set modules::instance::($this,times) [set ${namespace}::data(pollTimes)]}       ;# may not be available initially
            catch {set modules::instance::($this,views) [set ${namespace}::data(views)]}                  ;# there could be no views
        }

        proc synchronize {this} {
            module::synchronize $modules::instance::($this,loaded)
            initialize $this   ;# reinitialize in case a few variables have been set or reset in the module initialization procedure
        }

        proc empty {this} {
            module::clear $modules::instance::($this,loaded)
        }

    }


    set modules::(instances) {}

    proc modules {this} error                                                                                   ;# object-less class

    # using Tcl built-in package management facilities, seek available moodss modules
    proc available {{command {}}} {     ;# in command string, %M is replaced by module name and %S by switches list from module code
        catch {package require {}}                                              ;# make sure Tcl package auto loading search is done
        set packages {}
        foreach package [package names] {
            if {![info exists ::packageDirectory($package)]} continue                                        ;# for Tcl, for example
            set interpreter [interp create]              ;# use a separate interpreter in order not to interfere with loaded modules
            $interpreter eval "
                set auto_path [list $::auto_path]                      ;# copy a few required global variables in module interpreter
                set ::packageDirectory($package) $::packageDirectory($package)
            "
            if {\
                [catch {$interpreter eval "package require $package"}]||\
                ![$interpreter eval info exists ::${package}::data(updates)]\
            } continue                                                                                    ;# ignore invalid packages
            lappend packages $package
            set switches {}                                                                              ;# there may be no switches
            # module package name and module namespace are identical
            catch {set switches [$interpreter eval "set ::${package}::data(switches)"]}
            set switches [list $switches]                                                                    ;# make it a valid list
            interp delete $interpreter
            if {[string length $command]>0} {
                regsub -all %M $command $package string
                regsub -all %S $string $switches string
                uplevel #0 $string                                                          ;# always invoke command at global level
            }
        }
        return [lsort $packages]
    }

    proc printAvailable {} {            ;# using Tcl built-in package management facilities, seek and print available moodss modules
        puts {searching for module packages, please wait...}
        foreach package [available] {
            puts -nonewline "$package: possibly in"
            set count 0
            foreach directory $::auto_path {
                if {[file readable [file join $directory $package pkgIndex.tcl]]} {
                    if {$count>0} {
                        puts -nonewline ,
                    }
                    puts -nonewline " [file join $directory $package]"
                    incr count
                }
            }
            puts {}
        }
    }

    # recursive procedure: eventually initialize next module and its eventual options in command line arguments
    proc parse {arguments} {         ;# arguments list format is: module [-option [value] -option ...] module [-option ...]
        if {[llength $arguments]==0} return                                                                                  ;# done
        set name [lindex $arguments 0]
        set arguments [lrange $arguments 1 end]                                         ;# point to start of switches or next module
        foreach {name index} [decoded $name] {}  ;# eventually split module into its name and its index (if coming from a save file)
        if {![info exists ::packageDirectory($name)]} {                               ;# not a valid module (usually a wrong switch)
            error "error: \"$name\" is not a valid moodss module name"
        }
        if {![validName $name]} {
            error "\"$name\" module name contains invalid characters"
        }
        lifoLabel::push $global::messenger "loading $name..."
        update idletasks
        set instance [new instance $name $index]
        # load module since its switches need be known for command line parsing
        if {[catch {modules::instance::load $instance} message]} {
            if {$global::debug} {
                set information $::errorInfo
            }
            lifoLabel::pop $global::messenger
            delete $instance                                                                                             ;# clean up
            if {$global::debug} {
                error $information
            } else {
                error "module \"$name\" load error:\n$message"
            }
        }
        lifoLabel::pop $global::messenger
        set help [expr {[lsearch -exact $arguments --help]>=0}]                                         ;# help requested for module
        if {[info exists modules::instance::($instance,switches)]} {                                         ;# module takes options
            if {[llength $modules::instance::($instance,switches)]==0} {
                error "module \"$name\" switches are empty"             ;# design error: no need to recover when dynamically loading
            }
            if {$help} {
                displayHelpMessage $name $modules::instance::($instance,switches)
                exit
            }
            if {[catch\
                {set next [parseCommandLineArguments $modules::instance::($instance,switches) $arguments options]} message\
            ]} {
                delete $instance                                                                                         ;# clean up
                error "module \"$name\" options error: $message"
            }
            if {!$modules::instance::($instance,initialize)} {
                error "module \"$name\" has no initialize procedure"    ;# design error: no need to recover when dynamically loading
            }
            set modules::instance::($instance,options) [array get options]
            # save module arguments for eventual saving in file
            set modules::instance::($instance,arguments) [lrange $arguments 0 [expr {[llength $arguments]-[llength $next]-1}]]
            set arguments $next
        } else {                                                                                          ;# module takes no options
            if {$help} {
                displayHelpMessage $name
                exit
            }
            set modules::instance::($instance,arguments) {}                     ;# save module arguments for eventual saving in file
        }
        lappend modules::(instances) $instance                                     ;# add module to successfully loaded modules list
        parse $arguments                                                                                         ;# process the rest
        update idletasks                                       ;# make sure latest loading message is not left showing meaninglessly
    }

    proc helpHTMLData {name} {            ;# return HTML formatted help no matter whether provided plain or preformatted from module
        set displayArguments 0
        foreach instance $modules::(instances) {
            set namespace $modules::instance::($instance,namespace)
            foreach {module index} [decoded $namespace] {}
            if {[string compare $module $name]} continue
            # retrieve help text from first module namespace
            if {![info exists text]&&[catch {set ${namespace}::data(helpText)} text]} {
                set text {no help available}
            }
            if {![info exists header]} {
                set header "<h6>module $name</h6><i>version $modules::instance::($instance,version)"
            }
            lappend list $modules::instance::($instance,arguments)
            if {[string length $modules::instance::($instance,arguments)]>0} {
                set displayArguments 1
            }
        }
        if {$displayArguments} {
            append header {, invoked with instance arguments:<ol><b>}
            set index 0
            foreach arguments $list {
                append header "<li>:</b>  $arguments</li><b>"
            }
            append header </b></ol>
        }
        append header </i><br><br>
        if {[regsub -nocase <body> $text <body>$header text]==0} {                           ;# insert header if HTML formatted help
            regsub -all \n $text <br> text                                                 ;# regular help, keep original formatting
            return ${header}$text
        } else {                                                                                              ;# HTML formatted help
            return $text
        }
    }

    proc initialize {} {      ;# eventually invoke modules initialization procedure once. modules must be loaded first (see parse{})
        foreach instance $modules::(instances) {
            if {!$modules::instance::($instance,initialize)} continue
            set namespace $modules::instance::($instance,namespace)
            lifoLabel::push $global::messenger "initializing $namespace..."
            update idletasks
            set error 0
            if {[info exists modules::instance::($instance,options)]} {
                if {[catch {::${namespace}::initialize $instance::($instance,options)} message]} {
                    if {$global::debug} {
                        set information $::errorInfo
                    }
                    set error 1
                }
            } else {                                                                                      ;# module takes no options
                if {[catch ::${namespace}::initialize message]} {
                    if {$global::debug} {
                        set information $::errorInfo
                    }
                    set error 1
                }
            }
            lifoLabel::pop $global::messenger
            if {$error} {
                unload $instance                                                                                         ;# clean up
                if {$global::debug} {
                    error $information
                } else {
                    error "module \"$namespace\" initialize error:\n$message"
                }
            } else {
                modules::instance::synchronize $instance                                                 ;# in case data was updated
            }
            set modules::instance::($instance,initialize) 0                       ;# a module instance must be initialized once only
        }
        update idletasks                                ;# make sure latest initialization message is not left showing meaninglessly
    }

    proc setPollTimes {{override {}}} {
        if {[llength $modules::(instances)]==0} {
            set global::pollTimes {}
            set global::pollTime 0
            return
        }
        set default 0
        set minimum 0
        foreach instance $modules::(instances) {
            set times $modules::instance::($instance,times)
            if {[llength $times]==0} {
                error "module $modules::instance::($instance,namespace) poll times list is empty"
            }
            # for an asynchronous module, the sole time value would be negative and is used as graph interval, for example
            set time [lindex $times 0]
            if {$time<0} {                                          ;# asynchronous module, poll time is a viewer interval (negated)
                set intervals($time) {}
                continue
            }
            if {$time>$default} {                                                              ;# default value is the first in list
                set default $time                                                    ;# keep the greater default time of all modules
            }
            set times [lsort -integer $times]                                                                     ;# sort poll times
            set time [lindex $times 0]
            if {$time>$minimum} {
                set minimum $time                                                    ;# keep the greater minimum time of all modules
                set minimumModule $modules::instance::($instance,namespace)
            }
            foreach time $times {                                    ;# poll times list is the combination of all modules poll times
                set data($time) {}
            }
        }
        # sort and restrict poll times above maximum module minimum poll time
        set global::pollTimes [lsort -integer [array names data]]
        set global::pollTimes [lrange $global::pollTimes [lsearch -exact $global::pollTimes $minimum] end]
        set global::pollTime $default
        if {[string length $override]>0} {                                              ;# eventually validate command line override
            if {$override<$minimum} {
                puts stderr "$::argv0: minimum poll time is $minimum seconds for module $minimumModule"
                exit 1
            }
            set global::pollTime $override
        }
        if {$global::pollTime==0} { 
            # all modules are asynchronous, so use an average time as a viewer interval for viewers that need it, such as graphs.
            # the poll times list is empty at this point so the user cannot change the poll time.
            # note that the viewer interval can still be forced by the command line poll time option.
            set sum 0
            set number 0
            foreach interval [array names intervals] {
                incr sum $interval
                incr number
            }
            set global::pollTime [expr {round(double($sum)/-$number)}]
        }
    }

    proc identifier {array} {  ;# from an array name, eventually deduce a unique module identifier if needed (used in viewer labels)
        variable nextIndex

        set namespace [namespaceFromArray $array]
        foreach instance $modules::(instances) {
            if {[string compare $namespace $modules::instance::($instance,namespace)]==0} {                ;# this is a module array
                if {[info exists modules::instance::($instance,identifier)]} {
                    return $modules::instance::($instance,identifier)                               ;# favor identifier if it exists
                }
                foreach {module index} [decoded $namespace] {}
                if {[string length $index]>0} {        ;# there are more than 1 instance of this module, so identification is needed
                    return $namespace
                }
                break
            }
        }
        return {}                                                                   ;# not a module array or identification unneeded
    }

    proc decoded {name} {     ;# return module and index list (index may be empty if module is not indexed: name instead of name<N>)
        set index {}
        scan $name {%[^<]<%u>} name index                                     ;# eventually split module into its name and its index
        return [list $name $index]
    }

    proc validName {string} {                                                                               ;# includes iso-8859 set
        return [regexp {^[ 0-9a-zA-Z\240-\377~!@%^&*()_=+|;:',.?-]+$} $string]
    }

    proc displayHelpMessage {name {switches {}}} {
        puts -nonewline "$name module usage:"
        if {[llength $switches]==0} {
            puts -nonewline { <no arguments allowed>}
        } else {
            foreach {switch argument} $switches {
                puts -nonewline " \[$switch"
                if {$argument} {                                                                        ;# option takes one argument
                    puts -nonewline { argument}
                }
                puts -nonewline \]
            }
        }
        puts {}
    }

    proc loaded {} {    ;# return a list of namespaces with options, options being of list of switch, argument required and argument
        if {[llength $modules::(instances)]==0} {
            return {}
        }
        foreach instance $modules::(instances) {
            lappend list [list $instance $modules::instance::($instance,namespace)]
        }
        set return {}
        foreach list [lsort -dictionary -index 1 $list] {                                    ;# sort in namespace alphabetical order
            foreach {instance namespace} $list {}
            lappend return $namespace
            set switches {}                                                                       ;# in case module takes no options
            catch {set switches $modules::instance::($instance,switches)}
            if {[llength $switches]==0} {
                lappend return {}
            } else {
                catch {unset argument}
                foreach {switch required} $switches {
                    set argument($switch) $required
                }
                set arguments $modules::instance::($instance,arguments)
                set length [llength $arguments]
                set list {}
                for {set index 0} {$index<$length} {incr index} {
                    set item [lindex $arguments $index]
                    lappend list $item
                    if {$argument($item)} {
                        lappend list 1 [lindex $arguments [incr index]]                    ;# append value then point to next switch
                    } else {
                        lappend list 0 1                                  ;# value is true for boolean options since switch was used
                    }
                }
                lappend return $list
            }
        }
        return $return
    }

    proc names {} {  ;# return of list of all different loaded module names (a name is unique even if several instances were loaded)
        set list {}
        foreach instance $modules::(instances) {
            set module $modules::instance::($instance,module)
            if {[lsearch -exact $list $module]<0} {
                lappend list $module
            }
        }
        return $list
    }

### eventually remove ###
    proc namespaces {} {                                                 ;# return of list of all different loaded module namespaces
        set list {}
        foreach instance $modules::(instances) {
            lappend list $modules::instance::($instance,namespace)
        }
        return $list
    }

    proc unload {instance} {
        ldelete modules::(instances) $instance
        delete $instance
    }

    proc namespaceFromArray {name} {
        return [string trimleft [namespace qualifiers [namespace which -variable $name]] :]
    }

}
